"""Cancel All-Linking."""
from ..topics import CANCEL_ALL_LINKING
from .outbound_base import OutboundHandlerBase


# pylint: disable=arguments-differ
class CancelAllLinkingCommandHandler(OutboundHandlerBase):
    """Cancel All-Linking Command."""

    def __init__(self):
        """Init the CancelAllLinkingCommandHandler class."""
        super().__init__(topic=CANCEL_ALL_LINKING)



"""Dimmable Lighting Control Devices (CATEGORY 0x01)."""
from ..handlers.to_device.off import OffCommand
from ..handlers.to_device.off_fast import OffFastCommand
from ..handlers.to_device.on_fast import OnFastCommand
from ..handlers.to_device.on_level import OnLevelCommand
from ..handlers.to_device.status_request import StatusRequestCommand
from .device_commands import (
    OFF_COMMAND,
    OFF_FAST_COMMAND,
    ON_COMMAND,
    ON_FAST_COMMAND,
    STATUS_COMMAND,
)
from .open_close_controller_base import OpenCloseControllerBase


class OpenCloseResponderBase(OpenCloseControllerBase):
    """Variable Responder Base Device."""

    def open(self, open_level=0xFF, group=0, fast=False):
        """Open the device.

        Parameters:
            position: Default 0xff (full on). Set the device open level.
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        group = 1 if not group else group
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        self._handlers[group][command].send(on_level=open_level)

    async def async_open(self, open_level=0xFF, group=0, fast=False):
        """Open the device.

        Parameters:
            on_level: Default 0xff (full on). Set the device on level.
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        group = 1 if not group else group
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        return await self._handlers[group][command].async_send(on_level=open_level)

    def close(self, group=0, fast=False):
        """Close the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        group = 1 if not group else group
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        self._handlers[group][command].send()

    async def async_close(self, group=0, fast=False):
        """Close the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        group = 1 if not group else group
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        return await self._handlers[group][command].async_send()

    # pylint: disable=arguments-differ
    async def async_status(self):
        """Get the status of the device state."""
        return await self._handlers[STATUS_COMMAND].async_send()

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[STATUS_COMMAND] = StatusRequestCommand(self._address)
        group = 1
        if self._handlers.get(group) is None:
            self._handlers[group] = {}
        self._handlers[group][ON_COMMAND] = OnLevelCommand(self._address, group)
        self._handlers[group][OFF_COMMAND] = OffCommand(self._address, group)
        self._handlers[group][ON_FAST_COMMAND] = OnFastCommand(self._address, group)
        self._handlers[group][OFF_FAST_COMMAND] = OffFastCommand(self._address, group)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        group = 1
        state = self._groups[group]

        self._handlers[group][ON_COMMAND].subscribe(state.set_value)
        self._handlers[group][OFF_COMMAND].subscribe(state.set_value)
        self._handlers[group][ON_FAST_COMMAND].subscribe(state.set_value)
        self._handlers[group][OFF_FAST_COMMAND].subscribe(state.set_value)



"""Modem command to get the IM configuration."""
import logging

from pyinsteon.constants import ResponseStatus

from ..topics import GET_IM_CONFIGURATION
from . import ack_handler
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class GetImConfigurationHandler(OutboundHandlerBase):
    """Handle Get IM Configuration commands."""

    def __init__(self):
        """Init the GetImConfigurationHandler class."""
        super().__init__(topic=GET_IM_CONFIGURATION)

    # pylint: disable=arguments-differ
    @ack_handler
    async def async_handle_ack(
        self, disable_auto_linking, monitor_mode, auto_led, deadman
    ):
        """Receive the ACK message and return True."""
        await self._message_response.put(ResponseStatus.SUCCESS)
        self._call_subscribers(
            disable_auto_linking=disable_auto_linking,
            monitor_mode=monitor_mode,
            auto_led=auto_led,
            deadman=deadman,
        )



"""Manage inbound ON command from device."""
import logging

from ...constants import MessageFlagType
from ...topics import EXTENDED_GET_RESPONSE
from .. import inbound_handler
from ..inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class ExtendedGetResponseHandler(InboundHandlerBase):
    """Off Level command inbound."""

    def __init__(self, address):
        """Init the OffInbound class."""
        super().__init__(
            topic=EXTENDED_GET_RESPONSE,
            address=address,
            message_type=MessageFlagType.DIRECT,
        )

    @inbound_handler
    def handle_response(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the Extended Get response from a device."""
        data = {}
        for item in range(3, 15):
            data[f"data{item}"] = user_data[f"d{item}"]
        self._call_subscribers(group=user_data["d1"], data=data)



"""Handle sending a read request for ALDB records."""
import asyncio
import logging

from ... import pub
from ...address import Address
from ...constants import MessageFlagType, ResponseStatus
from ...topics import EXTENDED_READ_WRITE_ALDB, EXTENDED_READ_WRITE_ALDB_DIRECT_NAK
from ...utils import build_topic, subscribe_topic
from .. import direct_ack_handler, direct_nak_handler
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class ReadALDBCommandHandler(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=EXTENDED_READ_WRITE_ALDB, address=address)
        direct_nak_topic = build_topic(
            topic=EXTENDED_READ_WRITE_ALDB_DIRECT_NAK,
            address=self._address.id,
            message_type=MessageFlagType.DIRECT_NAK,
        )
        subscribe_topic(self.async_handle_direct_nak, direct_nak_topic)

    # pylint: disable=arguments-differ
    async def async_send(self, mem_addr: int = 0x0000, num_recs: int = 0):
        """Send ALDB read message asyncronously."""
        return await super().async_send(
            action=0x00, mem_addr=mem_addr, num_recs=num_recs
        )

    @direct_ack_handler
    async def async_handle_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left, topic=pub.AUTO_TOPIC
    ):
        """Handle the direct ACK."""
        # Need to make sure the ACK has time to aquire the lock
        await asyncio.sleep(0.05)
        if topic.name.endswith("direct_nak.direct_ack"):
            return
        if self._response_lock.locked():
            await self._direct_response.put(ResponseStatus.SUCCESS)
            self._update_subscribers_on_ack(cmd1, cmd2, target, user_data, hops_left)
            await asyncio.sleep(0.05)

    @direct_nak_handler
    async def async_handle_direct_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the message ACK."""
        # Need to make sure the ACK has time to aquire the lock
        await asyncio.sleep(0.05)
        if self._response_lock.locked():
            await self._direct_response.put(ResponseStatus.FAILURE)
            self._update_subscribers_on_nak(cmd1, cmd2, target, user_data, hops_left)

    def _update_subscribers_on_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Update subscribers."""
        self._call_subscribers(response=0)

    def _update_subscribers_on_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Update subscribers on DIIRECT NAK received."""
        self._call_subscribers(response=cmd2)



"""Insteon All-Link Database.

The All-Link database contains database records that represent links to other
Insteon devices that either respond to or control the current device.
"""
import logging

from ..constants import ALDBStatus, ALDBVersion
from ..managers.aldb_read_manager import ALDBReadManager
from .aldb_base import ALDBBase

_LOGGER = logging.getLogger(__name__)


class ALDB(ALDBBase):
    """All-Link Database for a device."""

    def __init__(
        self,
        address,
        version=ALDBVersion.V2,
        mem_addr=0x0FFF,
    ):
        """Init the ALDB class."""
        super().__init__(address=address, version=version, mem_addr=mem_addr)
        self._read_manager = ALDBReadManager(self)

    # pylint: disable=arguments-differ
    async def async_load(
        self, mem_addr: int = 0x00, num_recs: int = 0x00, refresh: bool = False
    ):
        """Load the All-Link Database."""
        _LOGGER.debug("Loading the ALDB async")
        self._update_status(ALDBStatus.LOADING)
        if refresh:
            self.clear()
        else:
            # Pop any unused records to make sure we query them
            unused = list(self.find(in_use=False))
            for rec in unused:
                self._records.pop(rec.mem_addr)
            self._hwm_record = None

        async for rec in self._read_manager.async_read(
            mem_addr=mem_addr, num_recs=num_recs
        ):
            # Make sure the records make sense
            if (
                self.high_water_mark_mem_addr
                and rec.mem_addr < self.high_water_mark_mem_addr
            ):
                continue

            # If an existing record will be replaced notify of change
            old_record = self._records.get(rec.mem_addr)
            if old_record and old_record.is_in_use:
                self._notify_change(self._records[rec.mem_addr], force_delete=True)

            self._records[rec.mem_addr] = rec
            if rec.is_in_use:
                self._notify_change(rec)
            if rec.is_high_water_mark:
                self._hwm_record = rec

            if self._calc_load_status():
                break

        self.set_load_status()

        return self._status



"""Handle sending a read request for ALDB records."""

from ...address import Address
from ...topics import EXTENDED_GET_SET_2
from .direct_command import DirectCommandHandlerBase


class ExtendedGet2Command(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=EXTENDED_GET_SET_2, address=address)
        self._data1 = 0x00

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send Get Operating Flags message asyncronously."""
        return await super().async_send(data1=self._data1)



"""Manage links beteween devices."""
from __future__ import annotations

import logging
from typing import Union

from ...address import Address
from ...constants import AllLinkMode, LinkStatus, ResponseStatus
from ...handlers.cancel_all_linking import CancelAllLinkingCommandHandler
from ...handlers.start_all_linking import StartAllLinkingCommandHandler
from ...handlers.to_device.enter_linking_mode import EnterLinkingModeCommand
from ...handlers.to_device.enter_unlinking_mode import EnterUnlinkingModeCommand

TIMEOUT = 3
_LOGGER = logging.getLogger(__name__)


async def async_enter_linking_mode(
    link_mode: AllLinkMode, group: int, address: Address = None
):
    """Put the Insteon Modem into linking mode."""
    link_cmd = StartAllLinkingCommandHandler()
    response = await link_cmd.async_send(link_mode=link_mode, group=group)
    _LOGGER.debug("Enter linking mode response: %s", str(response))

    if address is not None:
        link_cmd = EnterLinkingModeCommand(address)
        await link_cmd.async_send(group=group)
    return response


async def async_enter_unlinking_mode(group: int, address: Address = None):
    """Put the Insteon Modem into unlinking mode."""
    link_cmd = StartAllLinkingCommandHandler()
    link_mode = AllLinkMode.DELETE
    response = await link_cmd.async_send(link_mode=link_mode, group=group)

    if address is not None:
        link_cmd = EnterUnlinkingModeCommand(address)
        await link_cmd.async_send(group=group)

    return response


async def async_link_devices(
    controller, responder, group: int = 0, data1=255, data2=28, data3=1
):
    """Link two devices."""

    failed_1 = 0
    failed_2 = 0

    if not hasattr(controller, "address") or not hasattr(responder, "address"):
        raise TypeError("controller and responder must be devices")
    if _add_link_to_device(
        device=controller,
        target=responder.address,
        group=group,
        is_controller=True,
        data1=int(responder.cat),
        data2=responder.subcat,
        data3=responder.firmware,
    ):
        _, failed_1 = await controller.aldb.async_write()

    if _add_link_to_device(
        device=responder,
        target=controller.address,
        group=group,
        is_controller=False,
        data1=data1,
        data2=data2,
        data3=data3,
    ):
        _, failed_2 = await responder.aldb.async_write()

    if failed_1 or failed_2:
        return ResponseStatus.FAILURE
    return ResponseStatus.SUCCESS


async def async_cancel_linking_mode():
    """Cancel an All-Link session with the modem."""
    cmd = CancelAllLinkingCommandHandler()
    return await cmd.async_send()


async def async_unlink_devices(device1, device2, group: Union(int, None) = None):
    """Unlink two devices.

    `device2` can be passed as `Device` or `Address`. When `device2` are passed as `Address`, only the links in `device1` will be removed.

    `group` can be passed as an `int` or `None`. When `group` is `None` all links between the two devices will be removed. When
    `group` is an `int`, only that group wil be removed.
    """
    try:
        addr1 = device1.address
    except AttributeError:
        raise TypeError("device1 must be a Device")

    device2_is_device = True
    try:
        addr2 = device2.address
    except AttributeError:
        try:
            addr2 = Address(device2)
            device2_is_device = False
        except TypeError:
            raise TypeError("device2 must be a Device or an Address")

    failed_1 = 0
    failed_2 = 0

    for rec in device1.aldb.find(target=addr2, group=group):
        device1.aldb.modify(mem_addr=rec.mem_addr, in_use=False)
        _, failed_1 = await device1.aldb.async_write()

    if device2_is_device:
        for rec in device2.aldb.find(target=addr1, group=group):
            device2.aldb.modify(mem_addr=rec.mem_addr, in_use=False)
            _, failed_2 = await device2.aldb.async_write()

    if failed_1 or failed_2:
        return ResponseStatus.FAILURE
    return ResponseStatus.SUCCESS


def find_broken_links(devices):
    """Find proken links."""
    broken_link_list = {}
    for addr in devices:
        device = devices[addr]
        for mem_addr in device.aldb:
            rec = device.aldb[mem_addr]
            if rec.is_in_use:
                status = _test_broken(addr, rec, devices)
                if status != LinkStatus.FOUND:
                    if not broken_link_list.get(addr):
                        broken_link_list[addr] = {}
                    broken_link_list[addr][mem_addr] = (rec, status)
    return broken_link_list


def _test_broken(address, rec, devices):
    """Test if a corresponding record exists in liked device."""
    device = devices.get(rec.target)
    if not device:
        return LinkStatus.MISSING_TARGET

    if not device.aldb.is_loaded:
        return LinkStatus.TARGET_DB_NOT_LOADED

    for mem_addr in device.aldb:
        t_rec = device.aldb[mem_addr]
        if (
            t_rec.target == address
            and t_rec.group == rec.group
            and t_rec.is_controller != rec.is_controller
        ):
            return LinkStatus.FOUND
    if rec.is_controller:
        return LinkStatus.MISSING_RESPONDER
    return LinkStatus.MISSING_CONTROLLER


def _add_link_to_device(device, is_controller, group, target, data1, data2, data3):
    """Add a link to a device.

    If the link already exists it will not be added.
    """
    for rec in device.aldb.find(
        target=Address(target), group=group, is_controller=is_controller
    ):
        try:
            device.aldb.modify(
                mem_addr=rec.mem_addr,
                in_use=True,
                group=group,
                controller=is_controller,
                target=target,
                data1=data1,
                data2=data2,
                data3=data3,
            )
            return True
        except NotImplementedError:
            return False
    try:
        device.aldb.add(
            group=group,
            target=target,
            controller=is_controller,
            data1=data1,
            data2=data2,
            data3=data3,
        )
    except NotImplementedError:
        return False
    return True



"""Handle sending a read request for ALDB records."""

from ...address import Address
from ...topics import EXTENDED_GET_SET
from .. import ack_handler
from .direct_command import DirectCommandHandlerBase


class ThermostatGetSetPointCommand(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=EXTENDED_GET_SET, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send Get Operating Flags message asyncronously."""
        # This is not consistant with the 2441TH dev guide
        # It is consistand with 2441ZTH dev guide howerver
        response = await super().async_send(data3=0x01)
        return response

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle the ACK response.

        Required to ensure only GET requests are triggered.
        """
        if (
            not user_data
            or not user_data["d1"] == 0x00
            or not user_data["d2"] == 0x00
            or not user_data["d3"] == 0x01
        ):
            return
        await super().async_handle_ack(cmd1, cmd2, user_data)

    def _update_subscribers_on_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Update subscribers.

        Just need to notify listeners that the Set Point Response
        shoudl be coming.
        """
        self._call_subscribers()
